_base_ = [
    '../_base_/models/liteflownet2/liteflownet2_pre_M6S6.py',
    '../_base_/datasets/flyingchairs_320x448.py',
    '../_base_/schedules/liteflownet_pre_300k.py',
    '../_base_/default_runtime.py'
]
